##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Text index.
"""
from persistent import Persistent
from zope.interface import implementer

from zope.index.interfaces import IIndexSearch
from zope.index.interfaces import IInjection
from zope.index.interfaces import IStatistics
from zope.index.text.lexicon import CaseNormalizer
from zope.index.text.lexicon import Lexicon
from zope.index.text.lexicon import Splitter
from zope.index.text.lexicon import StopWordRemover
from zope.index.text.okapiindex import OkapiIndex
from zope.index.text.queryparser import QueryParser
import six

@implementer(IInjection, IIndexSearch, IStatistics)
class TextIndex(Persistent):
    """
    Text index.

    Implements :class:`zope.index.interfaces.IInjection` and
    :class:`zope.index.interfaces.IIndexSearch`.
    """

    def __init__(self, lexicon=None, index=None):
        """Provisional constructor.

        This creates the lexicon and index if not passed in.
        """
        _explicit_lexicon = True
        if lexicon is None:
            _explicit_lexicon = False
            lexicon = Lexicon(Splitter(), CaseNormalizer(), StopWordRemover())
        if index is None:
            index = OkapiIndex(lexicon)
        self.lexicon = _explicit_lexicon and lexicon or index.lexicon
        self.index = index

    def index_doc(self, docid, text):
        self.index.index_doc(docid, text)

    def unindex_doc(self, docid):
        self.index.unindex_doc(docid)

    def clear(self):
        self.index.clear()

    def documentCount(self):
        """Return the number of documents in the index."""
        return self.index.documentCount()

    def wordCount(self):
        """Return the number of words in the index."""
        return self.index.wordCount()

    def apply(self, querytext, start=0, count=None):
        parser = QueryParser(self.lexicon)
        tree = parser.parseQuery(querytext)
        results = tree.executeQuery(self.index)
        if results:
            qw = self.index.query_weight(tree.terms())

            # Hack to avoid ZeroDivisionError
            if qw == 0:
                qw = 1.0

            qw *= 1.0

            for docid, score in six.iteritems(results):
                try:
                    results[docid] = score/qw
                except TypeError:
                    # We overflowed the score, perhaps wildly unlikely.
                    # Who knows.
                    results[docid] = 2**64 // 10

        return results



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Topic Index tests
"""
import unittest

# pylint:disable=protected-access,blacklisted-name

_marker = object()

class TopicIndexTest(unittest.TestCase):

    def _getTargetClass(self):
        from zope.index.topic.index import TopicIndex
        return TopicIndex

    def _get_family(self):
        import BTrees
        return BTrees.family32

    def _makeOne(self, family=_marker, populate=False):
        if family is _marker:
            family = self._get_family()
        if family is None:
            index = self._getTargetClass()()
        index = self._getTargetClass()(family)
        if populate:
            self._addFilters(index)
            self._populate(index)
        return index

    def _search(self, index, query, expected, operator='and'):
        result = index.search(query, operator)
        self.assertEqual(result.keys(), expected)

    def _search_or(self, index, query, expected):
        return self._search(index, query, expected, 'or')

    def _search_and(self, index, query, expected):
        return self._search(index, query, expected, 'and')

    def _apply(self, index, query, expected, operator='and'):
        result = index.apply(query)
        self.assertEqual(result.keys(), expected)

    def _apply_or(self, index, query, expected):
        result = index.apply({'query': query, 'operator': 'or'})
        self.assertEqual(result.keys(), expected)

    def _apply_and(self, index, query, expected):
        result = index.apply({'query': query, 'operator': 'and'})
        self.assertEqual(result.keys(), expected)

    def test_class_conforms_to_IInjection(self):
        from zope.interface.verify import verifyClass
        from zope.index.interfaces import IInjection
        verifyClass(IInjection, self._getTargetClass())

    def test_instance_conforms_to_IInjection(self):
        from zope.interface.verify import verifyObject
        from zope.index.interfaces import IInjection
        verifyObject(IInjection, self._makeOne())

    def test_class_conforms_to_IIndexSearch(self):
        from zope.interface.verify import verifyClass
        from zope.index.interfaces import IIndexSearch
        verifyClass(IIndexSearch, self._getTargetClass())

    def test_instance_conforms_to_IIndexSearch(self):
        from zope.interface.verify import verifyObject
        from zope.index.interfaces import IIndexSearch
        verifyObject(IIndexSearch, self._makeOne())

    def test_class_conforms_to_ITopicQuerying(self):
        from zope.interface.verify import verifyClass
        from zope.index.topic.interfaces import ITopicQuerying
        verifyClass(ITopicQuerying, self._getTargetClass())

    def test_instance_conforms_to_ITopicQuerying(self):
        from zope.interface.verify import verifyObject
        from zope.index.topic.interfaces import ITopicQuerying
        verifyObject(ITopicQuerying, self._makeOne())

    def test_ctor_defaults(self):
        import BTrees
        index = self._makeOne(family=None)
        self.assertTrue(index.family is BTrees.family32)

    def test_ctor_explicit_family(self):
        import BTrees
        index = self._makeOne(family=BTrees.family64)
        self.assertTrue(index.family is BTrees.family64)

    def test_clear_erases_filters(self):
        index = self._makeOne()
        foo = DummyFilter('foo')
        index.addFilter(foo)
        index.clear()
        self.assertEqual(list(index._filters), [])

    def test_addFilter(self):
        index = self._makeOne()
        foo = DummyFilter('foo')
        index.addFilter(foo)
        self.assertEqual(list(index._filters), ['foo'])
        self.assertTrue(index._filters['foo'] is foo)

    def test_addFilter_duplicate_replaces(self):
        index = self._makeOne()
        foo = DummyFilter('foo')
        index.addFilter(foo)
        foo2 = DummyFilter('foo')
        index.addFilter(foo2)
        self.assertEqual(list(index._filters), ['foo'])
        self.assertTrue(index._filters['foo'] is foo2)

    def test_delFilter_nonesuch_raises_KeyError(self):
        index = self._makeOne()
        self.assertRaises(KeyError, index.delFilter, 'nonesuch')

    def test_delFilter(self):
        index = self._makeOne()
        foo = DummyFilter('foo')
        index.addFilter(foo)
        bar = DummyFilter('bar')
        index.addFilter(bar)
        index.delFilter('foo')
        self.assertEqual(list(index._filters), ['bar'])
        self.assertTrue(index._filters['bar'] is bar)

    def test_clearFilters_empty(self):
        index = self._makeOne()
        index.clearFilters() # doesn't raise

    def test_clearFilters_non_empty(self):
        index = self._makeOne()
        foo = DummyFilter('foo')
        index.addFilter(foo)
        bar = DummyFilter('bar')
        index.addFilter(bar)
        index.clearFilters()
        self.assertTrue(foo._cleared)
        self.assertTrue(bar._cleared)

    def test_index_doc(self):
        index = self._makeOne()
        foo = DummyFilter('foo')
        index.addFilter(foo)
        bar = DummyFilter('bar')
        index.addFilter(bar)
        obj = object()
        index.index_doc(1, obj)
        self.assertEqual(foo._indexed, [(1, obj)])
        self.assertEqual(bar._indexed, [(1, obj)])

    def test_unindex_doc(self):
        index = self._makeOne()
        foo = DummyFilter('foo')
        index.addFilter(foo)
        bar = DummyFilter('bar')
        index.addFilter(bar)
        index.unindex_doc(1)
        self.assertEqual(foo._unindexed, [1])
        self.assertEqual(bar._unindexed, [1])

    def test_search_non_tuple_list_query(self):
        index = self._makeOne()
        self.assertRaises(TypeError, index.search, {'nonesuch': 'ugh'})

    def test_search_bad_operator(self):
        index = self._makeOne()
        self.assertRaises(TypeError, index.search, ['whatever'], 'maybe')

    def test_search_no_filters_list_query(self):
        index = self._makeOne()
        result = index.search(['nonesuch'])
        self.assertEqual(set(result), set())

    def test_search_no_filters_tuple_query(self):
        index = self._makeOne()
        result = index.search(('nonesuch',))
        self.assertEqual(set(result), set())

    def test_search_no_filters_string_query(self):
        index = self._makeOne()
        result = index.search('nonesuch')
        self.assertEqual(set(result), set())

    def test_search_query_matches_one_filter(self):
        index = self._makeOne()
        foo = DummyFilter('foo', [1, 2, 3], self._get_family())
        index.addFilter(foo)
        bar = DummyFilter('bar', [2, 3, 4], self._get_family())
        index.addFilter(bar)
        result = index.search(['foo'])
        self.assertEqual(set(result), set([1, 2, 3]))

    def test_search_query_matches_multiple_implicit_operator(self):
        index = self._makeOne()
        foo = DummyFilter('foo', [1, 2, 3], self._get_family())
        index.addFilter(foo)
        bar = DummyFilter('bar', [2, 3, 4], self._get_family())
        index.addFilter(bar)
        result = index.search(['foo', 'bar'])
        self.assertEqual(set(result), set([2, 3]))

    def test_search_query_matches_multiple_implicit_op_no_intersect(self):
        index = self._makeOne()
        foo = DummyFilter('foo', [1, 2, 3], self._get_family())
        index.addFilter(foo)
        bar = DummyFilter('bar', [4, 5, 6], self._get_family())
        index.addFilter(bar)
        result = index.search(['foo', 'bar'])
        self.assertEqual(set(result), set())

    def test_search_query_matches_multiple_explicit_and(self):
        index = self._makeOne()
        foo = DummyFilter('foo', [1, 2, 3], self._get_family())
        index.addFilter(foo)
        bar = DummyFilter('bar', [2, 3, 4], self._get_family())
        index.addFilter(bar)
        result = index.search(['foo', 'bar'], operator='and')
        self.assertEqual(set(result), set([2, 3]))

    def test_search_query_matches_multiple_explicit_or(self):
        index = self._makeOne()
        foo = DummyFilter('foo', [1, 2, 3], self._get_family())
        index.addFilter(foo)
        bar = DummyFilter('bar', [2, 3, 4], self._get_family())
        index.addFilter(bar)
        result = index.search(['foo', 'bar'], operator='or')
        self.assertEqual(set(result), set([1, 2, 3, 4]))

    def test_apply_query_matches_multiple_non_dict_query(self):
        index = self._makeOne()
        foo = DummyFilter('foo', [1, 2, 3], self._get_family())
        index.addFilter(foo)
        bar = DummyFilter('bar', [2, 3, 4], self._get_family())
        index.addFilter(bar)
        result = index.apply(['foo', 'bar'])
        self.assertEqual(set(result), set([2, 3]))

    def test_apply_query_matches_multiple_implicit_op(self):
        index = self._makeOne()
        foo = DummyFilter('foo', [1, 2, 3], self._get_family())
        index.addFilter(foo)
        bar = DummyFilter('bar', [2, 3, 4], self._get_family())
        index.addFilter(bar)
        result = index.apply({'query': ['foo', 'bar']})
        self.assertEqual(set(result), set([2, 3]))

    def test_apply_query_matches_multiple_explicit_and(self):
        index = self._makeOne()
        foo = DummyFilter('foo', [1, 2, 3], self._get_family())
        index.addFilter(foo)
        bar = DummyFilter('bar', [2, 3, 4], self._get_family())
        index.addFilter(bar)
        result = index.apply({'query': ['foo', 'bar'], 'operator': 'and'})
        self.assertEqual(set(result), set([2, 3]))

    def test_apply_query_matches_multiple_explicit_or(self):
        index = self._makeOne()
        foo = DummyFilter('foo', [1, 2, 3], self._get_family())
        index.addFilter(foo)
        bar = DummyFilter('bar', [2, 3, 4], self._get_family())
        index.addFilter(bar)
        result = index.apply({'query': ['foo', 'bar'], 'operator': 'or'})
        self.assertEqual(set(result), set([1, 2, 3, 4]))

    def _addFilters(self, index):
        from zope.index.topic.filter import PythonFilteredSet
        index.addFilter(
            PythonFilteredSet('doc1', "context.meta_type == 'doc1'",
                              index.family))
        index.addFilter(
            PythonFilteredSet('doc2', "context.meta_type == 'doc2'",
                              index.family))
        index.addFilter(
            PythonFilteredSet('doc3', "context.meta_type == 'doc3'",
                              index.family))

    def _populate(self, index):

        class O(object):
            """ a dummy class """

            def __init__(self, meta_type):
                self.meta_type = meta_type

        index.index_doc(0, O('doc0'))
        index.index_doc(1, O('doc1'))
        index.index_doc(2, O('doc1'))
        index.index_doc(3, O('doc2'))
        index.index_doc(4, O('doc2'))
        index.index_doc(5, O('doc3'))
        index.index_doc(6, O('doc3'))

    def test_unindex(self):
        index = self._makeOne(populate=True)
        index.unindex_doc(-99)         # should not raise
        index.unindex_doc(3)
        index.unindex_doc(4)
        index.unindex_doc(5)
        self._search_or(index, 'doc1', [1, 2])
        self._search_or(index, 'doc2', [])
        self._search_or(index, 'doc3', [6])
        self._search_or(index, 'doc4', [])

    def test_or(self):
        index = self._makeOne(populate=True)
        self._search_or(index, 'doc1', [1, 2])
        self._search_or(index, ['doc1'], [1, 2])
        self._search_or(index, 'doc2', [3, 4])
        self._search_or(index, ['doc2'], [3, 4])
        self._search_or(index, ['doc1', 'doc2'], [1, 2, 3, 4])

    def test_and(self):
        index = self._makeOne(populate=True)
        self._search_and(index, 'doc1', [1, 2])
        self._search_and(index, ['doc1'], [1, 2])
        self._search_and(index, 'doc2', [3, 4])
        self._search_and(index, ['doc2'], [3, 4])
        self._search_and(index, ['doc1', 'doc2'], [])

    def test_apply_or(self):
        index = self._makeOne(populate=True)
        self._apply_or(index, 'doc1', [1, 2])
        self._apply_or(index, ['doc1'], [1, 2])
        self._apply_or(index, 'doc2', [3, 4])
        self._apply_or(index, ['doc2'], [3, 4])
        self._apply_or(index, ['doc1', 'doc2'], [1, 2, 3, 4])

    def test_apply_and(self):
        index = self._makeOne(populate=True)
        self._apply_and(index, 'doc1', [1, 2])
        self._apply_and(index, ['doc1'], [1, 2])
        self._apply_and(index, 'doc2', [3, 4])
        self._apply_and(index, ['doc2'], [3, 4])
        self._apply_and(index, ['doc1', 'doc2'], [])

    def test_apply(self):
        index = self._makeOne(populate=True)
        self._apply(index, 'doc1', [1, 2])
        self._apply(index, ['doc1'], [1, 2])
        self._apply(index, 'doc2', [3, 4])
        self._apply(index, ['doc2'], [3, 4])
        self._apply(index, ['doc1', 'doc2'], [])

class TopicIndexTest64(TopicIndexTest):

    def _get_family(self):
        import BTrees
        return BTrees.family64


class DummyFilter(object):

    _cleared = False

    def __init__(self, id, ids=(), family=None):
        self._id = id
        self._indexed = []
        self._unindexed = []
        self._family = family
        self._ids = ids

    def getId(self):
        return self._id

    def clear(self):
        self._cleared = True

    def index_doc(self, docid, obj):
        self._indexed.append((docid, obj))

    def unindex_doc(self, docid):
        self._unindexed.append(docid)

    def getIds(self):
        assert self._family is not None
        return self._family.IF.TreeSet(self._ids)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Text Index Tests
"""
import unittest

# pylint:disable=protected-access

_marker = object()

class TextIndexTests(unittest.TestCase):

    def _getTargetClass(self):
        from zope.index.text.textindex import TextIndex
        return TextIndex

    def _makeOne(self, lexicon=_marker, index=_marker):
        if lexicon is _marker:
            if index is _marker: # defaults
                return self._getTargetClass()()
            return self._getTargetClass()(index=index)
        else:
            if index is _marker:
                return self._getTargetClass()(lexicon)
            return self._getTargetClass()(lexicon, index)

    def test_class_conforms_to_IInjection(self):
        from zope.interface.verify import verifyClass
        from zope.index.interfaces import IInjection
        verifyClass(IInjection, self._getTargetClass())

    def test_instance_conforms_to_IInjection(self):
        from zope.interface.verify import verifyObject
        from zope.index.interfaces import IInjection
        verifyObject(IInjection, self._makeOne())

    def test_class_conforms_to_IIndexSearch(self):
        from zope.interface.verify import verifyClass
        from zope.index.interfaces import IIndexSearch
        verifyClass(IIndexSearch, self._getTargetClass())

    def test_instance_conforms_to_IIndexSearch(self):
        from zope.interface.verify import verifyObject
        from zope.index.interfaces import IIndexSearch
        verifyObject(IIndexSearch, self._makeOne())

    def test_class_conforms_to_IStatistics(self):
        from zope.interface.verify import verifyClass
        from zope.index.interfaces import IStatistics
        verifyClass(IStatistics, self._getTargetClass())

    def test_instance_conforms_to_IStatistics(self):
        from zope.interface.verify import verifyObject
        from zope.index.interfaces import IStatistics
        verifyObject(IStatistics, self._makeOne())

    def test_ctor_defaults(self):
        index = self._makeOne()
        from zope.index.text.lexicon import CaseNormalizer
        from zope.index.text.lexicon import Lexicon
        from zope.index.text.lexicon import Splitter
        from zope.index.text.lexicon import StopWordRemover
        from zope.index.text.okapiindex import OkapiIndex
        self.assertTrue(isinstance(index.index, OkapiIndex))
        self.assertTrue(isinstance(index.lexicon, Lexicon))
        self.assertTrue(index.index._lexicon is index.lexicon)
        pipeline = index.lexicon._pipeline
        self.assertEqual(len(pipeline), 3)
        self.assertTrue(isinstance(pipeline[0], Splitter))
        self.assertTrue(isinstance(pipeline[1], CaseNormalizer))
        self.assertTrue(isinstance(pipeline[2], StopWordRemover))

    def test_ctor_explicit_lexicon(self):
        from zope.index.text.okapiindex import OkapiIndex
        lexicon = object()
        index = self._makeOne(lexicon)
        self.assertTrue(index.lexicon is lexicon)
        self.assertTrue(isinstance(index.index, OkapiIndex))
        self.assertTrue(index.index._lexicon is lexicon)

    def test_ctor_explicit_index(self):
        lexicon = object()
        okapi = DummyOkapi(lexicon)
        index = self._makeOne(index=okapi)
        self.assertTrue(index.index is okapi)
        # See LP #232516
        self.assertTrue(index.lexicon is lexicon)

    def test_ctor_explicit_lexicon_and_index(self):
        lexicon = object()
        okapi = object()
        index = self._makeOne(lexicon, okapi)
        self.assertTrue(index.lexicon is lexicon)
        self.assertTrue(index.index is okapi)

    def test_index_doc(self):
        lexicon = object()
        okapi = DummyOkapi(lexicon)
        index = self._makeOne(lexicon, okapi)
        index.index_doc(1, 'cats and dogs')
        self.assertEqual(okapi._indexed[0], (1, 'cats and dogs'))

    def test_unindex_doc(self):
        lexicon = object()
        okapi = DummyOkapi(lexicon)
        index = self._makeOne(lexicon, okapi)
        index.unindex_doc(1)
        self.assertEqual(okapi._unindexed[0], 1)

    def test_clear(self):
        lexicon = object()
        okapi = DummyOkapi(lexicon)
        index = self._makeOne(lexicon, okapi)
        index.clear()
        self.assertTrue(okapi._cleared)

    def test_documentCount(self):
        lexicon = object()
        okapi = DummyOkapi(lexicon)
        index = self._makeOne(lexicon, okapi)
        self.assertEqual(index.documentCount(), 4)

    def test_wordCount(self):
        lexicon = object()
        okapi = DummyOkapi(lexicon)
        index = self._makeOne(lexicon, okapi)
        self.assertEqual(index.wordCount(), 45)

    def test_apply_no_results(self):
        lexicon = DummyLexicon()
        okapi = DummyOkapi(lexicon, {})
        index = self._makeOne(lexicon, okapi)
        self.assertEqual(index.apply('anything'), {})
        self.assertEqual(okapi._query_weighted, [])
        self.assertEqual(okapi._searched, ['anything'])

    def test_apply_w_results(self):
        lexicon = DummyLexicon()
        okapi = DummyOkapi(lexicon)
        index = self._makeOne(lexicon, okapi)
        results = index.apply('anything')
        self.assertEqual(results[1], 14.0 / 42.0)
        self.assertEqual(results[2], 7.4 / 42.0)
        self.assertEqual(results[3], 3.2 / 42.0)
        self.assertEqual(okapi._query_weighted[0], ['anything'])
        self.assertEqual(okapi._searched, ['anything'])

    def test_apply_w_results_zero_query_weight(self):
        lexicon = DummyLexicon()
        okapi = DummyOkapi(lexicon)
        okapi._query_weight = 0
        index = self._makeOne(lexicon, okapi)
        results = index.apply('anything')
        self.assertEqual(results[1], 14.0)
        self.assertEqual(results[2], 7.4)
        self.assertEqual(results[3], 3.2)
        self.assertEqual(okapi._query_weighted[0], ['anything'])
        self.assertEqual(okapi._searched, ['anything'])

    def test_apply_w_results_bogus_query_weight(self):
        DIVISOR = 2**64 // 10
        lexicon = DummyLexicon()
        # cause TypeError in division
        okapi = DummyOkapi(lexicon, {1: '14.0', 2: '7.4', 3: '3.2'})
        index = self._makeOne(lexicon, okapi)
        results = index.apply('anything')
        self.assertEqual(results[1], DIVISOR)
        self.assertEqual(results[2], DIVISOR)
        self.assertEqual(results[3], DIVISOR)
        self.assertEqual(okapi._query_weighted[0], ['anything'])
        self.assertEqual(okapi._searched, ['anything'])

class DummyOkapi(object):

    _cleared = False
    _document_count = 4
    _word_count = 45
    _query_weight = 42.0

    def __init__(self, lexicon, search_results=None):
        self.lexicon = lexicon
        self._indexed = []
        self._unindexed = []
        self._searched = []
        self._query_weighted = []
        if search_results is None:
            search_results = {1: 14.0, 2: 7.4, 3: 3.2}
        self._search_results = search_results

    def index_doc(self, docid, text):
        self._indexed.append((docid, text))

    def unindex_doc(self, docid):
        self._unindexed.append(docid)

    def clear(self):
        self._cleared = True

    def documentCount(self):
        return self._document_count

    def wordCount(self):
        return self._word_count

    def query_weight(self, terms):
        self._query_weighted.append(terms)
        return self._query_weight

    def search(self, term):
        self._searched.append(term)
        return self._search_results

    search_phrase = search_glob = search

class DummyLexicon(object):
    def parseTerms(self, term):
        return term



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Topic Index tests
"""
import unittest

_marker = object()

class FilteredSetBaseTests(unittest.TestCase):

    def _getTargetClass(self):
        from zope.index.topic.filter import FilteredSetBase
        return FilteredSetBase

    def _makeOne(self, id=None, expr=None, family=_marker):
        if id is None:
            id = 'test'
        if expr is None:
            expr = 'True'
        if family is _marker:
            return self._getTargetClass()(id, expr)
        return self._getTargetClass()(id, expr, family)

    def test_class_conforms_to_ITopicFilteredSet(self):
        from zope.interface.verify import verifyClass
        from zope.index.topic.interfaces import ITopicFilteredSet
        verifyClass(ITopicFilteredSet, self._getTargetClass())

    def test_instance_conforms_to_ITopicFilteredSet(self):
        from zope.interface.verify import verifyObject
        from zope.index.topic.interfaces import ITopicFilteredSet
        verifyObject(ITopicFilteredSet, self._makeOne())

    def test_ctor_defaults(self):
        import BTrees
        filter = self._makeOne(family=None)
        self.assertTrue(filter.family is BTrees.family32)
        self.assertEqual(filter.getId(), 'test')
        self.assertEqual(filter.getExpression(), 'True')
        self.assertEqual(len(filter.getIds()), 0)

    def test_ctor_explicit_family(self):
        import BTrees
        filter = self._makeOne(family=BTrees.family64)
        self.assertTrue(filter.family is BTrees.family64)

    def test_index_doc_raises_NotImplementedError(self):
        filter = self._makeOne()
        self.assertRaises(NotImplementedError, filter.index_doc, 1, object())

    def test_unindex_doc_missing_docid(self):
        filter = self._makeOne()
        filter.unindex_doc(1) # doesn't raise
        self.assertEqual(len(filter.getIds()), 0)

    def test_unindex_doc_existing_docid(self):
        filter = self._makeOne()
        filter._ids.insert(1)
        filter.unindex_doc(1)
        self.assertEqual(len(filter.getIds()), 0)

    def test_unindex_doc_existing_docid_w_residue(self):
        filter = self._makeOne()
        filter._ids.insert(1)
        filter._ids.insert(2)
        filter.unindex_doc(1)
        self.assertEqual(len(filter.getIds()), 1)

    def test_setExpression(self):
        filter = self._makeOne()
        filter.setExpression('False')
        self.assertEqual(filter.getExpression(), 'False')

class PythonFilteredSetTests(unittest.TestCase):

    def _getTargetClass(self):
        from zope.index.topic.filter import PythonFilteredSet
        return PythonFilteredSet

    def _makeOne(self, id=None, expr=None, family=_marker):
        if id is None:
            id = 'test'
        if expr is None:
            expr = 'True'
        return self._getTargetClass()(id, expr)

    def test_index_object_expr_True(self):
        filter = self._makeOne()
        filter.index_doc(1, object())
        self.assertEqual(list(filter.getIds()), [1])

    def test_index_object_expr_False(self):
        filter = self._makeOne(expr='False')
        filter.index_doc(1, object())
        self.assertEqual(len(filter.getIds()), 0)

    def test_index_object_expr_w_zero_divide_error(self):
        filter = self._makeOne(expr='1/0')
        filter.index_doc(1, object()) # doesn't raise
        self.assertEqual(len(filter.getIds()), 0)

def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(FilteredSetBaseTests),
        unittest.makeSuite(PythonFilteredSetTests),
    ))



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test zope.index.text.htmlsplitter
"""

from zope.index.text.tests.test_lexicon import SplitterTests

class HTMLWordSplitterTests(SplitterTests):

    def _getTargetClass(self):
        from zope.index.text.htmlsplitter import HTMLWordSplitter
        return HTMLWordSplitter

    def _makeOne(self):
        return self._getTargetClass()()

    def test_process_w_markup(self):
        splitter = self._makeOne()
        self.assertEqual(splitter.process(['<h1>abc</h1> &nbsp; <p>def</p>']),
                         ['abc', 'def'])

    def test_process_w_markup_no_spaces(self):
        splitter = self._makeOne()
        self.assertEqual(splitter.process(['<h1>abc</h1>&nbsp;<p>def</p>']),
                         ['abc', 'def'])

    def test_processGlob_w_markup_no_glob(self):
        splitter = self._makeOne()
        self.assertEqual(splitter.processGlob(['<h1>abc</h1> &nbsp; '
                                               '<p>def</p>']),
                         ['abc', 'def'])



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Keyword index
"""
import six
import BTrees
from BTrees.Length import Length
from persistent import Persistent


from zope.index.interfaces import IInjection, IStatistics, IIndexSearch
from zope.index.keyword.interfaces import IKeywordQuerying
from zope.interface import implementer


@implementer(IInjection, IStatistics, IIndexSearch, IKeywordQuerying)
class KeywordIndex(Persistent):
    """
    Keyword index.

    Implements :class:`zope.index.interfaces.IInjection`,
    :class:`zope.index.interfaces.IStatistics`,
    :class:`zope.index.interfaces.IIndexSearch` and
    :class:`zope.index.keyword.interfaces.IKeywordQuerying`.
    """

    family = BTrees.family32

    # If a word is referenced by at least tree_threshold docids,
    # use a TreeSet for that word instead of a Set.
    tree_threshold = 64

    def __init__(self, family=None):
        if family is not None:
            self.family = family
        self.clear()

    def clear(self):
        """Initialize forward and reverse mappings."""

        # The forward index maps index keywords to a sequence of docids
        self._fwd_index = self.family.OO.BTree()

        # The reverse index maps a docid to its keywords
        # TODO: Using a vocabulary might be the better choice to store
        # keywords since it would allow use to use integers instead of strings
        self._rev_index = self.family.IO.BTree()
        self._num_docs = Length(0)

    def documentCount(self):
        """Return the number of documents in the index."""
        return self._num_docs()

    def wordCount(self):
        """Return the number of indexed words"""
        return len(self._fwd_index)

    def has_doc(self, docid):
        return bool(docid in self._rev_index)

    def normalize(self, seq):
        """Perform normalization on sequence of keywords.

        Return normalized sequence. This method may be
        overriden by subclasses.

        """
        return seq

    def index_doc(self, docid, seq):
        if isinstance(seq, six.string_types):
            raise TypeError('seq argument must be a list/tuple of strings')

        old_kw = self._rev_index.get(docid, None)
        if not seq:
            if old_kw:
                self.unindex_doc(docid)
            return

        seq = self.normalize(seq)

        new_kw = self.family.OO.Set(seq)

        if old_kw is None:
            self._insert_forward(docid, new_kw)
            self._insert_reverse(docid, new_kw)
            self._num_docs.change(1)
        else:

            # determine added and removed keywords
            kw_added = self.family.OO.difference(new_kw, old_kw)
            kw_removed = self.family.OO.difference(old_kw, new_kw)

            # removed keywords are removed from the forward index
            for word in kw_removed:
                fwd = self._fwd_index[word]
                fwd.remove(docid)
                if not fwd:
                    del self._fwd_index[word]

            # now update reverse and forward indexes
            self._insert_forward(docid, kw_added)
            self._insert_reverse(docid, new_kw)

    def unindex_doc(self, docid):
        idx = self._fwd_index

        try:
            for word in self._rev_index[docid]:
                idx[word].remove(docid)
                if not idx[word]:
                    del idx[word]
        except KeyError:
            # 'WAAA!  Inconsistent'
            return

        try:
            del self._rev_index[docid]
        except KeyError: # pragma: no cover
            # 'WAAA!  Inconsistent'
            pass

        self._num_docs.change(-1)

    def _insert_forward(self, docid, words):
        """insert a sequence of words into the forward index """

        idx = self._fwd_index
        get_word_idx = idx.get
        IF = self.family.IF
        Set = IF.Set
        TreeSet = IF.TreeSet
        for word in words:
            word_idx = get_word_idx(word)
            if word_idx is None:
                idx[word] = word_idx = Set()
            word_idx.insert(docid)
            if (not isinstance(word_idx, TreeSet) and
                    len(word_idx) >= self.tree_threshold):
                # Convert to a TreeSet.
                idx[word] = TreeSet(word_idx)

    def _insert_reverse(self, docid, words):
        """ add words to forward index """

        if words:
            self._rev_index[docid] = words

    def search(self, query, operator='and'):
        """Execute a search given by 'query'."""
        if isinstance(query, six.string_types):
            query = [query]

        query = self.normalize(query)

        sets = []
        for word in query:
            docids = self._fwd_index.get(word, self.family.IF.Set())
            sets.append(docids)

        if operator == 'or':
            rs = self.family.IF.multiunion(sets)
        elif operator == 'and':
            # sort smallest to largest set so we intersect the smallest
            # number of document identifiers possible
            sets.sort(key=len)
            rs = None
            for set in sets:
                rs = self.family.IF.intersection(rs, set)
                if not rs:
                    break
        else:
            raise TypeError('Keyword index only supports `and` and `or` '
                            'operators, not `%s`.' % operator)

        if rs:
            return rs
        return self.family.IF.Set()

    def apply(self, query):
        operator = 'and'
        if isinstance(query, dict):
            if 'operator' in query:
                operator = query['operator']
            query = query['query']
        return self.search(query, operator=operator)

    def optimize(self):
        """Optimize the index. Call this after changing tree_threshold.

        This converts internal data structures between
        Sets and TreeSets based on tree_threshold.
        """
        idx = self._fwd_index
        IF = self.family.IF
        Set = IF.Set
        TreeSet = IF.TreeSet
        items = list(self._fwd_index.items())
        for word, word_idx in items:
            if len(word_idx) >= self.tree_threshold:
                if not isinstance(word_idx, TreeSet):
                    # Convert to a TreeSet.
                    idx[word] = TreeSet(word_idx)
            else:
                if isinstance(word_idx, TreeSet):
                    # Convert to a Set.
                    idx[word] = Set(word_idx)


class CaseInsensitiveKeywordIndex(KeywordIndex):
    """A case-normalizing keyword index (for strings as keywords)"""

    def normalize(self, seq):
        """
        Normalize by calling ``lower`` on every item in *seq*.
        """
        return [w.lower() for w in seq]
